class LoginError(Exception):
    def __init__(self, message='Login Error'):
        self.message = message
        super().__init__(self.message)

class AuthTimeOut(Exception):
    def __init__(self, message='Authentication timed out'):
        self.message = message
        super().__init__(self.message)


# from .curation import *

# from .curation import (SHFolderFormatter,
#                        STACGenerator,
#                        format_time_acquired,
#                        STACIdParser,
#                        StructuredParser,
#                        UnestructuredParser,
#                        STACDataFrame,
#                        read_stac)

# from .hello import say_hello, hola

# from .access import (SHClient,
#                      SHParametersFeature,
#                      EvalScripts,
#                      sentinel_1_download_parameters,
#                      sentinel_2_download_parameters,
#                      dem_download_parameters)
# from .tools import (stac_items_to_gdf,
#                     get_images_by_location,
#                     calculate_average_coordinates_distance,
#                     generate_new_locations_bounding_boxes,
#                     get_available_data_by_location)



from pydantic import BaseModel
import os
from pathlib import Path
import yaml
from ...models import Metadata


class IngestFolder:
    def __init__(self, repo, ingest_file, allowed_extensions, logger):
        self.repo = repo
        self.ingest_file = ingest_file
        self.allowed_extensions = allowed_extensions
        self.logger = logger if logger else print

    class Inputs(BaseModel):
        folder: Path
        user: dict
        force: bool = False
        delete: bool = False

    class Outputs(BaseModel):
        dataset: dict

    def __call__(self, inputs: Inputs) -> Outputs:
        # validate folder
        self.logger("Uploading directory (only files, not recursive)")
        items = list(inputs.folder.glob("*"))
        filtered_items = [item for item in items if item.is_file()]
        filtered_items = [
            item for item in filtered_items if item.suffix in self.allowed_extensions
        ]
        if len(filtered_items) == 0:
            raise Exception("No files found in directory")
        if len(filtered_items) > 10:
            raise Exception("Too many files in directory, limited to 10")
        if "metadata.yml" not in [item.name for item in filtered_items]:
            raise Exception("metadata.yml not found in directory")
        # load metadata
        metadata = yaml.safe_load(
            open(inputs.folder.joinpath("metadata.yml"), "r").read()
        )
        metadata = Metadata(**metadata)
        # remove metadata.yml from files
        filtered_items = [
            item for item in filtered_items if item.name != "metadata.yml"
        ]
        # create dataset
        data, error = self.repo.create_dataset(metadata.dict(), inputs.user["id_token"])
        # dataset may already exists, but if user is owner continue ingesting files
        current_files = []
        if error:
            data, error2 = self.repo.retrieve_dataset(metadata.name)
            if error2:
                raise Exception(error)
            if data["uid"] != inputs.user["sub"]:
                raise Exception("Dataset already exists.")
            data["dataset_id"] = data["id"]
            current_files = [item["name"] for item in data["files"]]
            if len(current_files) > 0 and not inputs.force:
                self.logger(
                    "The following files already exist and will not be uploaded (use --f to force re-upload):"
                )
                for item in current_files:
                    self.logger(f"{item}")
            # TODO: delete current_files that are not in filtered_items if --delete
            hanged_files = [
                file
                for file in current_files
                if file not in [item.name for item in filtered_items]
            ]
            if len(hanged_files) > 0:
                self.logger(
                    "The following files are no longer in your dataset (use --d to delete):"
                )
                for item in hanged_files:
                    self.logger(f"{item}")
                    if inputs.delete:
                        self.logger(f"Deleting file {item}...")
                        _, error = self.repo.delete_file(
                            data["dataset_id"], item, inputs.user["id_token"]
                        )
                        if error:
                            self.logger(error)
                        else:
                            self.logger("Done")
            filtered_items = [
                item for item in filtered_items if item.name not in current_files
            ]
        dataset_id = data["dataset_id"]
        # upload files
        if len(filtered_items) == 0:
            raise Exception("No files to upload")
        self.logger("The following files will be uploaded:")
        for item in filtered_items:
            self.logger(f"{item.name}")
        for item in filtered_items:
            data = self.ingest_file(item, dataset_id, logger=self.logger)
        return self.Outputs(dataset=data)



from ..src.repos import APIRepo
from ..src.usecases.datasets import DownloadDataset, DownloadFile
from .retrieve import retrieve_dataset
from ..auth import with_auth


@with_auth
def download_dataset(dataset, file, path=None, logger=None, user=None):
    api_repo = APIRepo()
    download = DownloadDataset(api_repo, retrieve_dataset, logger)
    inputs = download.Inputs(dataset=dataset, file=file, path=path, user=user)
    outputs = download(inputs)
    return outputs.dst_path



from pydantic import BaseModel
from typing import List


class RetrieveDatasets:
    def __init__(self, repo):
        self.repo = repo

    class Inputs(BaseModel):
        pass

    class Outputs(BaseModel):
        datasets: dict

    def __call__(self, inputs: Inputs) -> Outputs:
        data = self.repo.retrieve_datasets()
        datasets = {d["name"]: [f["name"] for f in d["files"]] for d in data}
        return self.Outputs(datasets=datasets)



"""
Module for formatter classes
"""

import rasterio

from os.path import join, exists
from os import listdir, mkdir
from glob import glob
from shutil import rmtree, copyfile
from .metadata import generate_raster_metadata


class Formatter:
    pass


class SHFolderFormatter(Formatter):
    """
    Class for formatting the directories of Sentinel images downloaded from Sentinel Hub Services
    """

    def __init__(self, root_folder):
        super().__init__()
        self.root = root_folder

    def format_folders(self) -> None:
        """
        Format the folder structure of the downloaded data from Sentinel Hub to a 
        digestible format, suitable for STAC metadata generation
        """
        sources = [dir for dir in listdir(self.root) if 'sen12floods' in dir]

        # Format the downloaded data directory structure
        for _source in sources:
            source = join(self.root, _source)   # abspath
            images_dirs = [image for image in listdir(source) if 'DS_Store' not in image]
            for _image in images_dirs:
                image_dir = join(source, _image)   # abspath
                split = _image.split('_')
                location_id = split[1]
                # There is no date for DEM data
                if len(split) >= 3:
                    date = split[2]
                    date_format = date.replace('-', '_')
                else:
                    date, date_format = None, None

                # Get response tiff
                _response_tiff = glob(f"{image_dir}/*/*.tiff")
                # If there is no response.tiff in the folder, skip it
                if not _response_tiff:
                    continue
                response_tiff = _response_tiff[0]

                # Create new destination folder
                dst_folder = join(source, f'{_source}_{location_id}')
                if date_format:
                    dst_folder = f'{dst_folder}_{date_format}'

                # If the destination folder already exists, skip it
                if exists(dst_folder):
                    continue
                mkdir(dst_folder)

                # Generate a metadata.json file of the raster in the destination folder. 
                # It will be used later for the STAC generation
                generate_raster_metadata(response_tiff, dst_folder, date)

                # Copy the response tiff to the destination folder giving a name 
                # indicating the mission
                if 'sentinel-1' in image_dir:
                    copyfile(response_tiff, join(dst_folder, 'SAR.tiff'))
                elif 'sentinel-2' in image_dir:
                    copyfile(response_tiff, join(dst_folder, 'SENTINEL-2.tiff'))
                elif 'dem' in image_dir:
                    copyfile(response_tiff, join(dst_folder, 'DEM.tiff'))

                # Remove the old image folder
                rmtree(image_dir)



from pydantic import BaseModel


class RetrieveDataset:
    def __init__(self, repo):
        self.repo = repo

    class Inputs(BaseModel):
        name: str

    class Outputs(BaseModel):
        dataset: dict

    def __call__(self, inputs: Inputs) -> Outputs:
        data, error = self.repo.retrieve_dataset(inputs.name)
        if error:
            raise Exception(error)
        return self.Outputs(dataset=data)



import typer
from pathlib import Path

from ..datasets import (
    retrieve_datasets,
    download_dataset,
    ingest_folder,
    ingest_stac,
)

app = typer.Typer()


@app.command()
def ingest(
    path: Path,
    f: bool = typer.Option(False, "--f", help="Force ingest even if file exists"),
    d: bool = typer.Option(False, "--d", help="Delete files not in the dataset"),
):
    """
    Ingest a dataset

    path: Path to folder with the dataset
    """
    try:
        if not path.is_dir():
            typer.echo("Path must be a folder")
            return
        if "catalog.json" in [f.name for f in path.iterdir()]:
            ingest_stac(str(path) + "/catalog.json", typer.echo)
        else:
            ingest_folder(path, f, d, typer.echo)
    except Exception as e:
        typer.echo(e)


@app.command()
def list():
    """
    List all datasets and files
    """
    datasets = retrieve_datasets()
    typer.echo(datasets)


@app.command()
def get(
    dataset: str,
    path: str = None,
    file: str = None,
):
    """
    Download a dataset

    dataset: Name of the dataset
    file: Name of the file to download (optional, if not provided, the whole dataset will be downloaded)
    path: Path to download the dataset to (optional, if not provided, the dataset will be downloaded to ~/.eotdl/datasets)
    """
    try:
        dst_path = download_dataset(dataset, file, path, typer.echo)
        typer.echo(f"Data available at {dst_path}")
    except Exception as e:
        typer.echo(e)


if __name__ == "__main__":
    app()



from ..src.repos import APIRepo
from ..src.usecases.datasets import RetrieveDatasets, RetrieveDataset


def list_datasets():
    return retrieve_datasets()


def retrieve_datasets():
    api_repo = APIRepo()
    retrieve = RetrieveDatasets(api_repo)
    inputs = retrieve.Inputs()
    outputs = retrieve(inputs)
    return outputs.datasets


def retrieve_dataset(name):
    api_repo = APIRepo()
    retrieve = RetrieveDataset(api_repo)
    inputs = retrieve.Inputs(name=name)
    outputs = retrieve(inputs)
    return outputs.dataset



from pydantic import BaseModel
from ....curation.stac import STACDataFrame
import json


class IngestSTAC:
    def __init__(self, repo, ingest_file, allowed_extensions):
        self.repo = repo
        self.ingest_file = ingest_file
        self.allowed_extensions = allowed_extensions

    class Inputs(BaseModel):
        stac_catalog: str
        dataset: str
        user: dict

    class Outputs(BaseModel):
        dataset: dict

    def __call__(self, inputs: Inputs) -> Outputs:
        # load the STAC catalog as a STACsetFrame
        df = STACDataFrame.from_stac_file(inputs.stac_catalog)
        # upload all assets to EOTDL
        for row in df.dropna(subset=["assets"]).iterrows():
            # for asset in df.assets.dropna().values[:10]:
            try:
                for k, v in row[1]["assets"].items():
                    data = self.ingest_file(
                        v["href"],
                        inputs.dataset,
                        allowed_extensions=self.allowed_extensions + [".tif", ".tiff"],
                    )
                    file_url = f"{self.repo.url}datasets/{data['dataset_id']}/download/{data['file_name']}"
                    df.loc[row[0], "assets"][k]["href"] = file_url
            except Exception as e:
                break
        data, error = self.repo.ingest_stac(
            json.loads(df.to_json()), inputs.dataset, inputs.user["id_token"]
        )
        if error:
            raise Exception(error)
        return self.Outputs(dataset=data)



from .stac import stac_items_to_gdf
from .sen12floods import (get_images_by_location,
                           calculate_average_coordinates_distance,
                           generate_new_locations_bounding_boxes,
                           get_available_data_by_location)


'''
Module for STAC parsers
'''

from os.path import dirname, basename


class STACIdParser:
    
    def get_item_id(self, raster_path: str):
        """
        Get the ID of the STAC Item from the given raster path

        :param raster_path: path to the raster file
        """
        pass


class StructuredParser(STACIdParser):

    def __init__(self) -> None:
        super().__init__()
    
    def get_item_id(self, raster_path: str):
        """
        Get the ID of the STAC Item from the given raster path.
        This function assumes that the project given by the user is structured,
        meaning that the raster files are stored in a folder with the same name

        :param raster_path: path to the raster file
        """
        tiff_dir_path = dirname(raster_path)
        id = tiff_dir_path.split('/')[-1]

        return id


class UnestructuredParser(STACIdParser):

    def __init__(self) -> None:
        super().__init__()
    
    def get_item_id(self, raster_path: str):
        """
        Get the ID of the STAC Item from the given raster path.
        This function assumes that the project given by the user is unstructured,
        meaning that the raster files are stored in the root folder or in a folder

        :param raster_path: path to the raster file
        """
        id = basename(raster_path).split('.')[0]

        return id



"""
Module for generating STAC metadata 
"""

import pandas as pd
import json
import pystac
from random import sample

from os import listdir
from os.path import join, basename, exists, dirname

import rasterio
from rasterio.warp import transform_bounds

from datetime import datetime
from shapely.geometry import Polygon, mapping
from glob import glob

from stac_validator.stac_validator import StacValidate

from .parsers import STACIdParser, StructuredParser
from .utils import format_time_acquired, count_ocurrences
from .extensions import type_stac_extensions_dict


class STACGenerator:
        
    def __init__(self, 
                 image_format: str='tiff',
                 catalog_type: pystac.CatalogType=pystac.CatalogType.SELF_CONTAINED, 
                 item_parser: STACIdParser=StructuredParser
                 ) -> None:
        """
        Initialize the STAC generator
        
        :param image_format: image format of the assets
        :param catalog_type: type of the catalog
        :param item_parser: parser to get the item ID
        """
        self._image_format = image_format
        self._catalog_type = catalog_type
        self._item_parser = item_parser()
        self._extensions_dict: dict = type_stac_extensions_dict
        self._validator = StacValidate(extensions=True)
        self._stac_dataframe = None

    def generate_stac_metadata(self,
                               stac_dataframe: pd.DataFrame,
                               id: str,
                               description: str,
                               output_folder: str='stac',
                               kwargs: dict={}) -> None:
        """
        Generate STAC metadata for a given directory containing the assets to generate metadata

        :param stac_dataframe: dataframe with the STAC metadata of a given directory containing the assets to generate metadata
        :param id: id of the catalog
        :param description: description of the catalog
        :param output_folder: output folder to write the catalog to
        """
        self._stac_dataframe = stac_dataframe
        
        # Create an empty catalog
        catalog = self.create_stac_catalog(id=id, description=description)
        
        # Add the collections to the catalog
        collections = self._stac_dataframe.collection.unique()
        for collection_path in collections:
            # TODO check if the items are directly under the root directory
            # Generate the collection
            collection = self.generate_stac_collection(collection_path)
            # Add the collection to the catalog
            catalog.add_child(collection)
        
        # Add the catalog to the root directory
        catalog.normalize_hrefs(output_folder)

        # Validate the catalog
        try:
            pystac.validation.validate(catalog)
            catalog.save(catalog_type=self._catalog_type)
        except pystac.STACValidationError as e:
            print(f'Catalog validation error: {e}')
            return

    def get_stac_dataframe(self, path: str, bands: dict=None, extensions: dict=None) -> pd.DataFrame:
        """
        Get a dataframe with the STAC metadata of a given directory containing the assets to generate metadata

        :param path: path to the root directory
        :param extensions: dictionary with the extensions
        :param image_format: image format of the assets
        """
        images = glob(str(path) + f'/**/*.{self._image_format}', recursive=True)
        images = sample(images, 50)   # TODO drop this line
        labels, ixs = self._format_labels(images)
        bands = self._get_items_list_from_dict(labels, bands)
        exts = self._get_items_list_from_dict(labels, extensions)
        collections = self._get_images_common_prefix(images)

        df = pd.DataFrame({'image': images, 
                           'label': labels, 
                           'ix': ixs, 
                           'collection': collections, 
                           'extensions': exts, 
                           'bands': bands})
        
        return df
    
    def _get_images_common_prefix(self, images: list) -> list:
        """
        Get the common prefix of a list of images

        :param images: list of images
        """
        images_common_prefix_dict = dict()

        images_dirs = [dirname(i) for i in images]

        for image in images_dirs:
            path = image
            common = False
            while not common:
                n = count_ocurrences(path, images_dirs)
                if n > 1:
                    images_common_prefix_dict[image] = path
                    common = True
                else:
                    path = dirname(path)

        images_common_prefix_list = list()
        for i in images:
            images_common_prefix_list.append(images_common_prefix_dict[dirname(i)])

        return images_common_prefix_list
    
    
    def _format_labels(self, images):
        """
        Format the labels of the images

        :param images: list of images
        """
        labels = [x.split('/')[-1].split('_')[0].split('.')[0] for x in images]
        ixs = [labels.index(x) for x in labels]
        return labels, ixs
    
    def _get_items_list_from_dict(self, labels: list, items: dict) -> list:
        """
        Get a list of items from a dictionary

        :param labels: list of labels
        :param items: dictionary with the items
        """
        if not items:
            # Create list of None with the same length as the labels list
            return [None for _ in labels]
        items_list = list()
        for label in labels:
            if label in items.keys():
                items_list.append(items[label])
            else:
                items_list.append(None)

        return items_list
    
    def _get_collection_extent(self, path: str) -> pystac.Extent:
        """
        Get the extent of a collection
        
        :param path: path to the directory
        """
        # Get the spatial extent of the collection
        spatial_extent = self._get_collection_spatial_extent(path)
        # Get the temporal interval of the collection
        temporal_interval = self._get_collection_temporal_interval(path)
        # Create the Extent object
        extent = pystac.Extent(spatial=spatial_extent, temporal=temporal_interval)

        return extent
    
    def _get_collection_spatial_extent(self, path: str) -> pystac.SpatialExtent:
        """
        Get the spatial extent of a collection

        :param path: path to the directory
        """
        # Get the bounding boxes of all the rasters in the path
        bboxes = list()
        # use glob
        rasters = glob(f'{path}/**/*.{self._image_format}', recursive=True)
        for raster in rasters:
            with rasterio.open(raster) as ds:
                bounds = ds.bounds
                dst_crs = 'EPSG:4326'
                try:
                    left, bottom, right, top = rasterio.warp.transform_bounds(ds.crs, dst_crs, *bounds)
                    bbox = [left, bottom, right, top]
                except rasterio.errors.CRSError:
                    spatial_extent = pystac.SpatialExtent([[0, 0, 0, 0]])
                    return spatial_extent
                bboxes.append(bbox)
        # Get the minimum and maximum values of the bounding boxes
        try:
            left = min([bbox[0] for bbox in bboxes])
            bottom = min([bbox[1] for bbox in bboxes])
            right = max([bbox[2] for bbox in bboxes])
            top = max([bbox[3] for bbox in bboxes])
            spatial_extent = pystac.SpatialExtent([[left, bottom, right, top]])
        except ValueError:
            spatial_extent = pystac.SpatialExtent([[0, 0, 0, 0]])
        finally:
            return spatial_extent
    
    def _get_collection_temporal_interval(self, path: str) -> pystac.TemporalExtent:
        """
        Get the temporal interval of a collection

        :param path: path to the directory
        """
        # Get all the metadata.json files in the path
        metadata_json_files = glob(f'{path}/**/*.json', recursive=True)
        if not metadata_json_files:
            return self._get_unknow_temporal_interval()
        
        # Get the temporal interval of every metadata.json file
        temporal_intervals = list()
        for metadata_json_file in metadata_json_files:
            with open(metadata_json_file, 'r') as f:
                metadata = json.load(f)
            temporal_intervals.append(metadata['date-adquired']) if metadata['date-adquired'] else None
        if temporal_intervals:   # TODO control in DEM data
            try:
                # Get the minimum and maximum values of the temporal intervals
                min_date = min([datetime.strptime(interval, '%Y-%m-%d') for interval in temporal_intervals])
                max_date = max([datetime.strptime(interval, '%Y-%m-%d') for interval in temporal_intervals])
            except ValueError:
                min_date = datetime.strptime('2000-01-01', '%Y-%m-%d')
                max_date = datetime.strptime('2023-12-31', '%Y-%m-%d')
            finally:
                # Create the temporal interval
                temporal_interval = pystac.TemporalExtent([min_date, max_date])
        else:
            return self._get_unknow_temporal_interval()

        return temporal_interval
    
    def _get_unknow_temporal_interval(self) -> pystac.TemporalExtent:
        """
        Get an unknown temporal interval
        """
        min_date = datetime.strptime('2000-01-01', '%Y-%m-%d')
        max_date = datetime.strptime('2023-12-31', '%Y-%m-%d')

        return pystac.TemporalExtent([min_date, max_date])

    def create_stac_catalog(self, id: str, description: str, kwargs: dict={}) -> pystac.Catalog:
        """
        Create a STAC catalog

        :param id: id of the catalog
        :param description: description of the catalog
        :param params: additional parameters
        """
        return pystac.Catalog(id=id, description=description, **kwargs)

    def generate_stac_collection(self, path: str) -> pystac.Collection:
        """
        Generate a STAC collection from a directory containing the assets to generate metadata

        :param path: path to the root directory
        """
        # Get the collection extent
        extent = self._get_collection_extent(path)
        # Create the collection
        collection = pystac.Collection(id=basename(path),
                                        description='Collection',
                                        extent=extent)
        
        for image in self._stac_dataframe.image:
            # Check if the path of the image is a child of the path of the collection
            if path in image:
                # Create the item
                item = self.create_stac_item(image)
                # Add the item to the collection
                collection.add_item(item)
        
        # Return the collection
        return collection

    def create_stac_collection(self, id: str, description: str, extent: pystac.Extent, kwargs: dict={}) -> pystac.Collection:
        """
        Create a STAC collection

        :param id: id of the collection
        :param description: description of the collection
        :param extent: extent of the collection
        :param params: additional parameters
        """
        return pystac.Collection(id=id, description=description, extent=extent, **kwargs)

    def create_stac_item(self,
                        raster_path: str,
                        kwargs: dict={}
                        ) -> pystac.Item:
        """
        Create a STAC item from a directory containing the raster files and the metadata.json file

        :param raster_path: path to the raster file
        """
        # Check if there is any metadata file in the directory associated to the raster file
        metadata = self._get_item_metadata(raster_path)

        # Obtain the bounding box from the raster
        with rasterio.open(raster_path) as ds:
            bounds = ds.bounds
            dst_crs = 'EPSG:4326'
            try:
                left, bottom, right, top = rasterio.warp.transform_bounds(ds.crs, dst_crs, *bounds)
            except rasterio.errors.CRSError:
                # If the raster has no crs, set the bounding box to 0
                left, bottom, right, top = 0, 0, 0, 0

        # Create bbox
        bbox = [left, bottom, right, top]

        # Create geojson feature
        # If the bounding box has no values, set the geometry to None
        geom = mapping(Polygon([
        [left, bottom],
        [left, top],
        [right, top],
        [right, bottom]
        ]))

        # Initialize properties
        properties = dict()

        # Obtain the date acquired
        if metadata and metadata["date-adquired"]:
            time_acquired = format_time_acquired(metadata["date-adquired"])
        else:
            # Set unknown date
            time_acquired = datetime.strptime('2000-01-01', '%Y-%m-%d')
        
        # Obtain the item ID. The approach depends on the item parser
        id = self._item_parser.get_item_id(raster_path)
        
        # Instantiate pystac item
        item = pystac.Item(id=id,
                geometry=geom,
                bbox=bbox,
                datetime=time_acquired,
                properties=properties,
                **kwargs)
        
        # Get the item extension using the dataframe, from the raster path
        extensions = self._stac_dataframe[self._stac_dataframe['image'] == raster_path]['extensions'].values
        extensions = extensions[0] if extensions else None
        # Add the required extensions to the item
        if extensions:
            if isinstance(extensions, str):
                extensions = [extensions]
            for extension in extensions:
                extension_obj = self._extensions_dict[extension]
                extension_obj.add_extension_to_object(item)

        # Add the assets to the item
        # First of all, we need to get the image bands and extract them from the raster
        # in order to create the assets
        bands = self._stac_dataframe[self._stac_dataframe['image'] == raster_path]['bands'].values
        bands = bands[0] if bands else None
        if not bands:
            # If there is no bands, create a single band asset from the file, assuming thats a singleband raster
            href = basename(raster_path)
            title = basename(raster_path).split('.')[0]
            asset = pystac.Asset(href=href, title=title, media_type=pystac.MediaType.GEOTIFF)
        else:
            with rasterio.open(raster_path, 'r') as raster:
                # Get the name of the raster file without extension
                raster_name = basename(raster_path).split('.')[0]
                if isinstance(bands, str):
                    bands = [bands]
                for band in bands:
                    i = bands.index(band)
                    try:
                        single_band = raster.read(i + 1)
                    except IndexError:
                        # TODO put try here for IndexError: band index 2 out of range (not in (1,))
                        # TODO control
                        single_band = raster.read(1)
                    band_name = f'{raster_name}_{band}.{self._image_format}'
                    output_band = join(dirname(raster_path), band_name)
                    # Copy the metadata
                    metadata = raster.meta.copy()
                    metadata.update({"count": 1})
                    # Write the band to the output folder
                    with rasterio.open(output_band, "w", **metadata) as dest:
                        dest.write(single_band, 1)
                    # Instantiate pystac asset
                    asset = pystac.Asset(href=band_name, title=band, media_type=pystac.MediaType.GEOTIFF)
                    # Add the asset to the item
                    item.add_asset(band_name, asset)
                    # Add the required extensions to the asset if required
                    if extensions:
                        if isinstance(extensions, str):
                            extensions = [extensions]
                        for extension in extensions:
                            extension_obj = self._extensions_dict[extension]
                            extension_obj.add_extension_to_object(asset)

        
        return item

    def _get_item_metadata(self, raster_path: str) -> str:
        """
        Get the metadata JSON file of a given directory, associated to a raster file

        :param raster_path: path to the raster file
        """
        # Get the directory of the raster file
        raster_dir_path = dirname(raster_path)
        # Get the metadata JSON file
        # Check if there is a metadata.json file in the directory
        if 'metadata.json' in listdir(raster_dir_path):
            metadata_json = join(raster_dir_path, 'metadata.json')
        else:
            # If there is no metadata.json file in the directory, check if there is
            # a json file with the same name as the raster file
            raster_name = raster_path.split('/')[-1]
            raster_name = raster_name.split('.')[0]
            metadata_json = join(raster_dir_path, f'{raster_name}.json')
            if not exists(metadata_json):
                # If there is no metadata.json file in the directory, return None
                return None
        
        # Open the metadata.json file and return it
        with open(metadata_json, 'r') as f:
            metadata = json.load(f)
        
        return metadata



import requests
from tqdm import tqdm
from pathlib import Path
import os
from concurrent.futures import ThreadPoolExecutor
import time
import multiprocessing
import hashlib
import geopandas as gpd


class APIRepo:
    def __init__(self, url=os.getenv("EOTDL_API_URL", "https://api.eotdl.com/")):
        self.url = url
        # print(self.url)

    def login(self):
        return requests.get(self.url + "auth/login")

    def token(self, code):
        return requests.get(self.url + "auth/token?code=" + code)

    def logout_url(self):
        response = requests.get(self.url + "auth/logout")
        return response.json()["logout_url"]

    def create_dataset(self, metadata, id_token):
        response = requests.post(
            self.url + "datasets",
            json=metadata,
            headers={"Authorization": "Bearer " + id_token},
        )
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]

    def retrieve_datasets(self):
        return requests.get(self.url + "datasets").json()

    def retrieve_dataset(self, name):
        response = requests.get(self.url + "datasets?name=" + name)
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]

    def download_file(self, dataset, dataset_id, file, id_token, path):
        url = self.url + "datasets/" + dataset_id + "/download/" + file
        headers = {"Authorization": "Bearer " + id_token}
        path = f"{path}/{file}"
        with requests.get(url, headers=headers, stream=True) as r:
            r.raise_for_status()
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024 * 1024 * 10
            progress_bar = tqdm(
                total=total_size, unit="iB", unit_scale=True, unit_divisor=1024
            )
            with open(path, "wb") as f:
                for chunk in r.iter_content(block_size):
                    progress_bar.update(len(chunk))
                    if chunk:
                        f.write(chunk)
            progress_bar.close()
            return path

    def ingest_file(self, file, dataset_id, id_token, checksum=None):
        reponse = requests.post(
            self.url + "datasets/" + dataset_id,
            files={"file": open(file, "rb")},
            data={"checksum": checksum} if checksum else None,
            headers={"Authorization": "Bearer " + id_token},
        )
        if reponse.status_code != 200:
            return None, reponse.json()["detail"]
        return reponse.json(), None

    def ingest_file_url(self, file, dataset, id_token):
        reponse = requests.post(
            self.url + "datasets/url",
            json={"dataset": dataset, "url": file},
            headers={"Authorization": "Bearer " + id_token},
        )
        if reponse.status_code != 200:
            return None, reponse.json()["detail"]
        return reponse.json(), None

    def read_in_chunks(self, file_object, CHUNK_SIZE):
        while True:
            data = file_object.read(CHUNK_SIZE)
            if not data:
                break
            yield data

    def prepare_large_upload(self, file, dataset_id, checksum, id_token):
        filename = Path(file).name
        response = requests.post(
            self.url + f"datasets/{dataset_id}/uploadId",
            json={"name": filename, "checksum": checksum},
            headers={"Authorization": "Bearer " + id_token},
        )
        if response.status_code != 200:
            raise Exception(response.json()["detail"])
        data = response.json()
        upload_id, parts = (
            data["upload_id"],
            data["parts"] if "parts" in data else [],
        )
        return upload_id, parts

    def get_chunk_size(self, content_size):
        # adapt chunk size to content size to avoid S3 limits (10000 parts, 500MB per part, 5TB per object)
        chunk_size = 1024 * 1024 * 10  # 10 MB (up to 100 GB, 10000 parts)
        if content_size >= 1024 * 1024 * 1024 * 100:  # 100 GB
            chunk_size = 1024 * 1024 * 100  # 100 MB (up to 1 TB, 10000 parts)
        elif content_size >= 1024 * 1024 * 1024 * 1000:  # 1 TB
            chunk_size = 1024 * 1024 * 500  # 0.5 GB (up to 5 TB, 10000 parts)
        return chunk_size

    def ingest_large_dataset(self, file, upload_id, id_token, parts):
        content_path = os.path.abspath(file)
        content_size = os.stat(content_path).st_size
        chunk_size = self.get_chunk_size(content_size)
        total_chunks = content_size // chunk_size
        # upload chunks sequentially
        pbar = tqdm(
            self.read_in_chunks(open(content_path, "rb"), chunk_size),
            total=total_chunks,
        )
        index = 0
        for chunk in pbar:
            part = index // chunk_size + 1
            offset = index + len(chunk)
            index = offset
            if part not in parts:
                checksum = hashlib.md5(chunk).hexdigest()
                response = requests.post(
                    self.url + "datasets/chunk/" + upload_id,
                    files={"file": chunk},
                    data={"part_number": part, "checksum": checksum},
                    headers={"Authorization": "Bearer " + id_token},
                )
                if response.status_code != 200:
                    raise Exception(response.json()["detail"])
            pbar.set_description(
                "{:.2f}/{:.2f} MB".format(
                    offset / 1024 / 1024, content_size / 1024 / 1024
                )
            )
        pbar.close()
        return

    def complete_upload(self, id_token, upload_id):
        r = requests.post(
            self.url + "datasets/complete/" + upload_id,
            headers={"Authorization": "Bearer " + id_token},
        )
        if r.status_code != 200:
            return None, r.json()["detail"]
        return r.json(), None

    def update_dataset(self, name, path, id_token, checksum):
        # check that dataset exists
        data, error = self.retrieve_dataset(name)
        if error:
            return None, error
        # first call to get upload id
        dataset_id = data["id"]
        url = self.url + f"datasets/chunk/{dataset_id}?checksum={checksum}"
        response = requests.get(url, headers={"Authorization": "Bearer " + id_token})
        if response.status_code != 200:
            return None, response.json()["detail"]
        data = response.json()
        _, upload_id, parts = data["dataset_id"], data["upload_id"], data["parts"]
        # assert dataset_id is None
        content_path = os.path.abspath(path)
        content_size = os.stat(content_path).st_size
        url = self.url + "datasets/chunk"
        chunk_size = 1024 * 1024 * 100  # 100 MiB
        total_chunks = content_size // chunk_size
        headers = {
            "Authorization": "Bearer " + id_token,
            "Upload-Id": upload_id,
            "Dataset-Id": dataset_id,
        }
        # upload chunks sequentially
        pbar = tqdm(
            self.read_in_chunks(open(content_path, "rb"), chunk_size),
            total=total_chunks,
        )
        index = 0
        for chunk in pbar:
            offset = index + len(chunk)
            part = index // chunk_size + 1
            index = offset
            if part not in parts:
                headers["Part-Number"] = str(part)
                file = {"file": chunk}
                r = requests.post(url, files=file, headers=headers)
                if r.status_code != 200:
                    return None, r.json()["detail"]
            pbar.set_description(
                "{:.2f}/{:.2f} MB".format(
                    offset / 1024 / 1024, content_size / 1024 / 1024
                )
            )
        pbar.close()
        # complete upload
        url = self.url + "datasets/complete"
        r = requests.post(
            url,
            json={"checksum": checksum},
            headers={
                "Authorization": "Bearer " + id_token,
                "Upload-Id": upload_id,
                "Dataset-Id": dataset_id,
            },
        )
        if r.status_code != 200:
            return None, r.json()["detail"]
        return r.json(), None

    def delete_file(self, dataset_id, file_name, id_token):
        response = requests.delete(
            self.url + "datasets/" + dataset_id + "/file/" + file_name,
            headers={"Authorization": "Bearer " + id_token},
        )
        if response.status_code != 200:
            return None, response.json()["detail"]
        return response.json(), None

    def ingest_stac(self, stac_json, dataset, id_token):
        reponse = requests.post(
            self.url + "datasets/stac",
            json={"dataset": dataset, "stac": stac_json},
            headers={"Authorization": "Bearer " + id_token},
        )
        if reponse.status_code != 200:
            return None, reponse.json()["detail"]
        return reponse.json(), None

    def download_stac(self, dataset_id, id_token):
        url = self.url + "datasets/" + dataset_id + "/download"
        headers = {"Authorization": "Bearer " + id_token}
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            return None, response.json()["detail"]
        return gpd.GeoDataFrame.from_features(response.json()["features"]), None



"""
Module for data engineering in the sen12floods dataset
"""
from ...access.sentinelhub.client import SHClient
from ...access.sentinelhub.utils import (sentinel_1_search_parameters, 
                                         sentinel_2_search_parameters)
from statistics import mean

import geopandas as gpd
from shapely import geometry


def get_images_by_location(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Generate a GeoDataFrame with the available images for each location in the dataset. 

    :param gdf: GeoDataFrame generated from the ItemCollection of a sen12floods collection
    :return gdf_dates_per_aoi:  GeoDataFrame with the available images for each location in
                                the dataset.
            The returned GeoDataFrame has three columns:
                - location_id: the unique ID of each location.
                - images_count: the count of available images of each location.
                - images_dates: list with the dates of the available images of each location.
    """
    uniques_location_id = gdf['location_id'].unique()   # List of unique location ids
    uniques_location_id.sort()

    images_count_list, images_dates_list = [], []

    # Iterate the unique location ids, count the number of images per location and generate
    # a list with the dates of every image in a location
    for location_id in uniques_location_id:
        dates = gdf[gdf['location_id'] == location_id]['datetime']
        images_count_list.append(dates.count())
        images_dates_list.append(dates.tolist())

    data = {'sequence_id': uniques_location_id, 'dates_count': images_count_list, 'dates_list': images_dates_list}
    gdf_dates_per_aoi = gpd.GeoDataFrame.from_dict(data)

    return gdf_dates_per_aoi


def calculate_average_coordinates_distance(bounding_box_by_location: dict) -> list:
    """
    Calculate the mean distance between maximum and minixum longitude and latitude of the bounding boxes
    from the existing locations. This is intended to use these mean distance to generate the bounding 
    boxes of the new locations given a centroid.

    :param bounding_box_by_location: dictionary with format location_id : bounding_box for the existing
            locations in the sen12floods dataset.
    :return mean_long_diff, mean_lat_diff: mean longitude and latitude difference in the bounding boxes
    """
    long_diff_list, lat_diff_list = list(), list()

    for bbox in bounding_box_by_location.values():
        long_diff = bbox[2] - bbox[0]
        long_diff_list.append(long_diff)
        lat_diff = bbox[3] - bbox[1]
        lat_diff_list.append(lat_diff)

    mean_long_diff = mean(long_diff_list)
    mean_lat_diff = mean(lat_diff_list)

    return mean_long_diff, mean_lat_diff

def generate_bounding_box(geom: geometry.point.Point, 
                          differences: list
                          ) -> list:
    """
    Generate the bounding box of a given point using the difference
    between the maximum and mininum coordinates of the bounding box

    :param geom: shapely geometry object of the point which we want to
                generate the bounding box.
    :param differences: list with the difference between the maximum
                and minimum longitude and latitude coordinates.
    :return: list with the resulting bounding box from the computing.
    """
    long_diff, lat_diff = differences[0], differences[1]
    lon, lat = geom.x, geom.y
    
    bbox = (lon - (long_diff/2), 
            lat - (lat_diff/2), 
            lon + (long_diff/2), 
            lat + (lat_diff/2))
    
    # Round the coordinates to 6 decimals
    bounding_box = [round(i, 6) for i in bbox]

    return bounding_box


def generate_new_locations_bounding_boxes(gdf: gpd.GeoDataFrame,
                                          mean_differences: list,
                                          latest_id: int
                                          ) -> dict:
    """
    Generate the bounding box of every new location, using the mean difference between the maximum and
    minimum calculated longitude and latitude. This function also returns the time interval which we
    want to request from Sentinel Hub Services.

    :param gdf: GeoDataFrame wiht the new locations that are going to be added to the dataset
    :param mean_differences: list with the longitude and latitude mean differences, which are going to be used
            to generate the bounding boxes.
    :return: bbox_by_new_location: dict with format {<location_id>: {'bounding_box': list(), 'time_interval': list()}, ... }
            that contains the bounding box and time interval of the imagery for each location
    """
    bbox_by_new_location = dict()

    for i, row in gdf.iterrows():
        new_location_id = str(latest_id + 1)
        time_interval = row['Began'].strftime("%Y-%m-%d"), row['Ended'].strftime("%Y-%m-%d")
        bbox = generate_bounding_box(row['geometry'], mean_differences)
        bbox_by_new_location[new_location_id] = {'bounding_box': bbox, 'time_interval': time_interval}
        latest_id += 1

    return bbox_by_new_location


sentinel_parameters = {'sentinel-1': sentinel_1_search_parameters,
                       'sentinel-2': sentinel_2_search_parameters}


def get_available_data_by_location(search_data: dict,
                                   eotdl_client: SHClient,
                                   sentinel_mission: str
                                   ) -> list:
    """
    Search and return a dict with the available Sentinel data for a dict with given locations and a time intervals.

    :param search_data: dictionary with the data required to search the available imagery in a given location
            and time interval. It must have the following format:
                {<location_id>: {'bounding_box': list(), 'time_interval': list()}, ... }
    :param eotdl_client: eotdl.SHClient object required to search for availabe data in Sentinel Hub 
    :param sentinel_mission: id of the required Sentinel mission. The value must be <sentinel-1> or <sentinel-2>
    :return: available_data: available data for downloading for a given location and time interval
    :return: not_available_data: list with the locations that does not have any available data for the
            given location and time interval
    """
    if sentinel_mission not in ('sentinel-1', 'sentinel-2'):
        print('The specified Sentinel mission is not valid. The values must be between <sentinel-1> and <sentinel-2>')
        return
    
    parameters = sentinel_parameters[sentinel_mission]

    available_data, not_available_data = dict(), list()
    for location_id, location_info in search_data.items():
        parameters.bounding_box = location_info['bounding_box']
        parameters.time_interval = location_info['time_interval']
        results = eotdl_client.search_available_sentinel_data(parameters)
        if results:
            # The returning results are composed by a list with format 
            # 'id': <image ID>, properties : {'datetime': <image date>}
            # As we can't make a bulk request with the ID but with the date time,
            # and we need all the available images in a time lapse and not
            # a mosaic, we are going to generate a dict with format
            # 'location_id': <location ID>,
            # {'bounding_box': <image bbox>, 'time_interval': <image date>}
            # This dictionary is digerible by the SHClient
            time_intervals = list()
            for result in results:
                datetime = result['properties']['datetime'][0:10]
                time_interval = (datetime, datetime)
                time_intervals.append(time_interval) if time_interval not in time_intervals else time_intervals
            available_data[location_id] = {'bounding_box': location_info['bounding_box'], 'time_interval': time_intervals}
        else:
            # We should have a trace with the locations without
            # available data
            not_available_data.append(location_id)

    return available_data, not_available_data



"""
Module for EO-TDL metadata
"""

import datetime
import json
import rasterio
from rasterio.warp import transform_bounds

from os.path import dirname, join, exists


def generate_raster_metadata(raster_path: str,
                             output_folder: str,
                             date_adquired: str|datetime.datetime
                             ) -> None:
    """
    Generate metadata.json file for a raster file

    :param raster_path: path to the raster file
    :param output_folder: output folder to write the metadata.json file to
    :param date_adquired: date adquired of the raster file
    """
    with rasterio.open(raster_path) as ds:
        bounds = ds.bounds
        dst_crs = 'EPSG:4326'  # EPSG identifier for WGS84 coordinate system used by the geojson format
        left, bottom, right, top = rasterio.warp.transform_bounds(ds.crs, dst_crs, *bounds)
        bbox = [left, bottom, right, top]

    # Get raster directory path to get the request.json file
    raster_dir_path = dirname(raster_path)

    # Read the request.json file and get the request data type
    if exists(raster_dir_path):
        with open(join(raster_dir_path, 'request.json'), 'r') as f:
            request = json.load(f)
            request_data_type = request['request']['payload']['input']['data'][0]['type']

    metadata_path = join(output_folder, 'metadata.json')
    metadata = {'date-adquired': date_adquired, 'bounding-box': bbox, 'type': request_data_type}
    
    with open(metadata_path, 'w') as f:
        json.dump(metadata, f)



from pydantic import BaseModel
import os
import typing

from ....src.utils import calculate_checksum


class IngestFile:
    def __init__(self, repo, allowed_extensions, logger):
        self.repo = repo
        self.allowed_extensions = allowed_extensions
        self.logger = logger if logger else print

    class Inputs(BaseModel):
        file: typing.Any
        dataset_id: str
        user: dict

    class Outputs(BaseModel):
        data: dict

    def __call__(self, inputs: Inputs) -> Outputs:
        # validate file extension
        extension = os.path.splitext(inputs.file)[1]
        if extension not in self.allowed_extensions:
            raise Exception(
                f"Only {', '.join(self.allowed_extensions)} files are allowed"
            )
        id_token = inputs.user["id_token"]
        self.logger(f"Uploading file {inputs.file}...")
        # if inputs.file.startswith("http://") or inputs.file.startswith("https://"):
        #     data, error = self.repo.ingest_file_url(
        #         inputs.file, inputs.metadata.name, id_token
        #     )
        # else:
        self.logger("Computing checksum...")
        checksum = calculate_checksum(inputs.file)
        self.logger(checksum)
        self.logger("Ingesting file...")
        filesize = os.path.getsize(inputs.file)
        # ingest small file
        if filesize < 1024 * 1024 * 16:  # 16 MB
            data, error = self.repo.ingest_file(
                inputs.file, inputs.dataset_id, id_token, checksum
            )
            if error:
                raise Exception(error)
            self.logger("Done")
            return self.Outputs(data=data)
        # ingest large file
        upload_id, parts = self.repo.prepare_large_upload(
            inputs.file, inputs.dataset_id, checksum, id_token
        )
        self.repo.ingest_large_dataset(inputs.file, upload_id, id_token, parts)
        self.logger("\nCompleting upload...")
        data, error = self.repo.complete_upload(id_token, upload_id)
        if error:
            raise Exception(error)
        self.logger("Done")
        return self.Outputs(data=data)



from .sentinelhub import (SHClient, 
                          SHParametersFeature, 
                          EvalScripts, 
                          sentinel_1_download_parameters, 
                          sentinel_2_download_parameters,
                          dem_download_parameters)


from pydantic import BaseModel

class Logout:
    def __init__(self, repo, api_repo):
        self.repo = repo
        self.api_repo = api_repo

    class Inputs(BaseModel):
        pass 

    class Outputs(BaseModel):
        logout_url: str

    def __call__(self, inputs: Inputs) -> Outputs:
        self.repo.logout()
        logout_url = self.api_repo.logout_url()
        return self.Outputs(logout_url=logout_url)


from pydantic import BaseModel
from ....src.utils import calculate_checksum
from ....curation.stac import STACDataFrame
from pathlib import Path
import os


class DownloadDataset:
    def __init__(self, repo, retrieve_dataset, logger):
        self.repo = repo
        self.retrieve_dataset = retrieve_dataset
        self.logger = logger if logger else print

    class Inputs(BaseModel):
        dataset: str
        file: str = None
        path: str = None
        user: dict

    class Outputs(BaseModel):
        dst_path: str

    def download(self, dataset, dataset_id, file, checksum, path, user):
        self.logger(f"Downloading {file}")
        dst_path = self.repo.download_file(
            dataset, dataset_id, file, user["id_token"], path
        )
        if calculate_checksum(dst_path) != checksum:
            self.logger(f"Checksum for {file} does not match")
        self.logger(f"Done")
        return dst_path

    def __call__(self, inputs: Inputs) -> Outputs:
        dataset = self.retrieve_dataset(inputs.dataset)
        if inputs.path is None:
            download_path = str(Path.home()) + "/.eotdl/datasets/" + inputs.dataset
        else:
            download_path = inputs.path + "/" + inputs.dataset
        os.makedirs(download_path, exist_ok=True)
        if dataset["quality"] == 0:
            if inputs.file:
                files = [f for f in dataset["files"] if f["name"] == inputs.file]
                if not files:
                    raise Exception(f"File {inputs.file} not found")
                if len(files) > 1:
                    raise Exception(f"Multiple files with name {inputs.file} found")
                dst_path = self.download(
                    inputs.dataset,
                    dataset["id"],
                    inputs.file,
                    files[0]["checksum"],
                    download_path,
                    inputs.user,
                )
                return self.Outputs(dst_path=dst_path)
            for file in dataset["files"]:
                dst_path = self.download(
                    inputs.dataset,
                    dataset["id"],
                    file["name"],
                    file["checksum"],
                    download_path,
                    inputs.user,
                )
            return self.Outputs(dst_path="/".join(dst_path.split("/")[:-1]))
        else:
            gdf, error = self.repo.download_stac(
                dataset["id"],
                inputs.user["id_token"],
            )
            if error:
                raise Exception(error)
            df = STACDataFrame(gdf)
            # df.geometry = df.geometry.apply(lambda x: Polygon() if x is None else x)
            path = inputs.path
            if path is None:
                path = str(Path.home()) + "/.eotdl/datasets/" + dataset["name"]
            df.to_stac(path)
            return self.Outputs(dst_path=path)



import os

from ..src.repos import APIRepo
from ..src.usecases.datasets import IngestFile, IngestFolder, IngestSTAC
from ..auth import with_auth


allowed_extensions = [
    ".zip",
    ".tar",
    ".tar.gz",
    ".csv",
    ".txt",
    ".json",
    ".pdf",
    ".md",
    ".yml",
]


def ingest_q1(dataset, stac_catalog):
    print("hola")
    return


@with_auth
def ingest_file(
    file, dataset_id, logger=None, allowed_extensions=allowed_extensions, user=None
):
    api_repo = APIRepo()
    ingest = IngestFile(api_repo, allowed_extensions, logger)
    inputs = ingest.Inputs(file=file, dataset_id=dataset_id, user=user)
    outputs = ingest(inputs)
    return outputs.data


@with_auth
def ingest_folder(folder, force, delete, logger=None, user=None):
    api_repo = APIRepo()
    ingest = IngestFolder(api_repo, ingest_file, allowed_extensions, logger)
    inputs = ingest.Inputs(folder=folder, user=user, force=force, delete=delete)
    outputs = ingest(inputs)
    return outputs.dataset


@with_auth
def ingest_stac(stac_catalog, dataset, logger=None, user=None):
    api_repo = APIRepo()
    ingest = IngestSTAC(api_repo, ingest_file, allowed_extensions)
    inputs = ingest.Inputs(stac_catalog=stac_catalog, dataset=dataset, user=user)
    outputs = ingest(inputs)
    return outputs.dataset



"""
Module for the STAC dataframe
"""

import pandas as pd
import geopandas as gpd
import pystac
import json
from geomet import wkt
from os.path import join
from os import makedirs
from typing import Union
from math import isnan
from .utils import convert_df_geom_to_shape, get_all_children


class STACDataFrame(gpd.GeoDataFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_stac_file(self, stac_file):
        """
        Create a STACDataFrame from a STAC file
        """
        return read_stac(stac_file)

    def to_stac(self, path):
        """
        Create a STAC catalog and children from a STACDataFrame
        """
        df = self.copy()

        if "id" in df.columns and "stac_id" in df.columns:
            id_column = "stac_id"
            stac_id_exists = True
        else:
            id_column = "id"
            stac_id_exists = False

        # First, create the catalog and its folder, if exists
        catalog_df = df[df["type"] == "Catalog"]

        if catalog_df.empty:
            makedirs(path, exist_ok=True)
        else:
            for index, row in catalog_df.iterrows():
                root_output_folder = path + "/" + row[id_column]
                makedirs(root_output_folder, exist_ok=True)
                row_json = row.to_dict()

                # Curate the json row
                row_json = self.curate_json_row(row_json, stac_id_exists)

                with open(join(root_output_folder, f"catalog.json"), "w") as f:
                    json.dump(row_json, f)

        # Second, create the collections and their folders, if exist
        collections = dict()
        collections_df = df[df["type"] == "Collection"]
        for index, row in collections_df.iterrows():
            stac_output_folder = join(root_output_folder, row[id_column])
            collections[row[id_column]] = stac_output_folder
            makedirs(stac_output_folder, exist_ok=True)
            row_json = row.to_dict()

            # Curate the json row
            row_json = self.curate_json_row(row_json, stac_id_exists)

            with open(join(stac_output_folder, f"collection.json"), "w") as f:
                json.dump(row_json, f)

        # Then, create the items and their folders, if exist
        features_df = df[df["type"] == "Feature"]
        for index, row in features_df.iterrows():
            collection = row["collection"]
            stac_output_folder = join(collections[collection], row[id_column])

            # Convert the geometry from WKT back to geojson
            row["geometry"] = row["geometry"].wkt
            row["geometry"] = wkt.loads(row["geometry"])
            makedirs(stac_output_folder, exist_ok=True)
            row_json = row.to_dict()

            # Curate the json row
            row_json = self.curate_json_row(row_json, stac_id_exists)

            with open(join(stac_output_folder, f'{row_json["id"]}.json'), "w") as f:
                json.dump(row_json, f)

    def curate_json_row(self, row: dict, stac_id_exists: bool) -> dict:
        """
        Curate the json row of a STACDataFrame, in order to generate a valid STAC file

        :param row: row of a STACDataFrame
        :param stac_id_exists: if the stac_id column exists
        """
        keys_to_remove = list()

        # Remove the created_at and modified_at columns, if the STACDataFrame comes from GeoDB
        for i in "created_at", "modified_at":
            if i in row.keys():
                keys_to_remove.append(i)

        # Rename the stac_id column to id, to avoid conflicts with the id column
        if stac_id_exists:
            row["id"] = row["stac_id"]
            del row["stac_id"]

        # Remove the NaN values and empty strings
        for k, v in row.items():
            if (isinstance(v, float) and isnan(v)) or v == "":
                keys_to_remove.append(k)
        for key in keys_to_remove:
            del row[key]
        del row["geometry"]

        return row


def read_stac(
    stac_file: Union[pystac.Catalog, pystac.Collection, str],
    geometry_column: str = "geometry",
) -> STACDataFrame:
    """
    Read a STAC file and return a STACDataFrame

    :param stac_file: STAC file to read
    :param geometry_column: name of the geometry column
    """
    if isinstance(stac_file, str):
        stac_file = pystac.read_file(stac_file)
    children = get_all_children(stac_file)

    # Convert Dataframe to STACDataFrame
    dataframe = pd.DataFrame(children)
    dataframe[geometry_column] = dataframe.apply(convert_df_geom_to_shape, axis=1)
    stac_dataframe = STACDataFrame(
        dataframe,
        crs="EPSG:4326",
        geometry=gpd.GeoSeries.from_wkt(dataframe[geometry_column]),
    )

    return stac_dataframe
